#ifndef FWDPY11_TSEVOLUTION_REMOVE_EXTINCT_MUTATIONS_HPP
#define FWDPY11_TSEVOLUTION_REMOVE_EXTINCT_MUTATIONS_HPP

#include <fwdpy11/types/Population.hpp>

void
remove_extinct_mutations(fwdpy11::Population& pop);

#endif

