/*!
  \file forward_types.hpp
  \defgroup basicTypes Mutation and genome data types
  These are the basic types common to all simulations: mutations and genome.
  See @ref md_md_policies for more details.
*/
#ifndef _FORWARD_TYPES_HPP_
#define _FORWARD_TYPES_HPP_

#include <fwdpp/fundamental_types/typedefs.hpp>
#include <fwdpp/fundamental_types/mutation_base.hpp>
#include <fwdpp/fundamental_types/haploid_genome.hpp>

#endif
