#ifndef FWDPP_UTIL_NESTED_FORWARD_LISTS_HPP
#define FWDPP_UTIL_NESTED_FORWARD_LISTS_HPP

#warning("Deprecated: use <fwdpp/types/nested_forward_lists.hpp> instead")

#include <fwdpp/types/nested_forward_lists.hpp>

#endif
