#ifndef FWDPP_TS_SIMPLIFICATION_FLAGS_HPP
#define FWDPP_TS_SIMPLIFICATION_FLAGS_HPP

#include <cstdint>
#include <type_traits>

namespace fwdpp
{
    namespace ts
    {
        enum class simplification_flags : std::uint32_t
        {
        };
    }
}
#endif
