#ifndef __FWDPP_SUGAR_MUTATION_POPGENMUT_HPP__
#define __FWDPP_SUGAR_MUTATION_POPGENMUT_HPP__

#warning "This header is deprecated.  Please use fwdpp/types/mutation.hpp."

#include <fwdpp/types/mutation.hpp>

#endif
